/*	Editor.java
     
   Copyright 2007 Manfred Kopp
   
    This software distributed under the terms of the 
    GNU General Public License Version 3 (GPLv3) of the License, 
    or (at your option) any later version.
        
    This file is part of the game Step-By-Step.

    Step-By-Step is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Step-By-Step is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.step_by_step;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile; 

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.Timer;
import javax.swing.filechooser.FileFilter;


public class Editor extends Playfield {
	static final long serialVersionUID= 1L;

	protected boolean trailStarted;
	protected boolean isStartCell; // no flags able to set on first cell
	protected boolean teleportStarted;
	protected int teleportNr;
	private JFileChooser fc;
	protected String solution, compactSolution;

	public Editor() {
		System.out.println("Constructor Editor");
		fc = new JFileChooser();
		fc.addChoosableFileFilter(new LevelFileFilter());
		fc.setAcceptAllFileFilterUsed(false);
		moveTimer= new Timer(1, this);	
	}

	protected void resetField() {
		int x,y;
		trailStarted= false;
		isStartCell= true; 
		lastMove= 0;

		startX= F_SIZE_X / 2;
		startY= F_SIZE_Y / 2;		
		for (y= 0; y< F_SIZE_Y; y++) {
			for (x= 0; x< F_SIZE_X; x++) {
				actualField[y][x]= F_EMPTY;
			}
		}
		posX= startX;
		posY= startY;
		nextNr= 0;   // 1 is the only selectable number-flag
		levelNr=0;
		showLevelNr= levelNr;
		level[0].nrTeleports=0;
		for (int i=0; i<MAX_TELEPORTS;i++) {
			showTeleport[i]=false;			
		}
		cellsLeft= Integer.MAX_VALUE;		// 
		teleportStarted= false;
		teleportNr=0;
		solution="";
	}

	protected void loadImages() {
		super.loadImages();
		try {
			bImg[F_EMPTY]= ImageIO.read( new File 
				("gfx","editor-empty.gif"));
		} catch (Exception e) {
			bImg[F_EMPTY]=null;
		} 
	}

	protected void loadLevels() {
	}

	public void keyPressed(KeyEvent e) {
		handleKeyCode( e.getKeyCode());
	}

	protected void handleKeyCode(int c) {	

		// debug
		System.out.printf("Keycode:%d\n", c);
		if (c == KeyEvent.VK_N) {
			System.out.println("Key N");
		}

		// if game is stoped or animation running , only a reset is possible
		if (isKeyReleased && c == KeyEvent.VK_N) {
			resetLevel();
			repaint();
		}

		// if an animation is running (waiting for next timer event)
		// no input is accepted
		if (!waitForInput) {
			return;
		}

		if (isKeyReleased && !gameStoped) {
			int flag= actualField[posY][posX] & F_FLAG_MASK;
			int type= actualField[posY][posX] & F_TYPE_MASK;

			if (teleportStarted) {
				if (c == KeyEvent.VK_T) {
					Teleport t;
					t= level[0].teleport[teleportNr];
					if ((t.fromX == t.toX && t.fromY == t.toY) || flag!=0) {
						// do not make a teleport 
						showTeleport[teleportNr]= false;
						posX= level[0].teleport[teleportNr].fromX;
						posY= level[0].teleport[teleportNr].fromY;
						actualField[posY][posX] &= F_TYPE_MASK; // clear teleport-flag 
						level[0].nrTeleports--;
					}
					else {
						teleportNr++;
						updateField();
						// temporyry set teleport-up flag, so no other flags can
						// be set on this cell
						actualField[posY][posX]= (byte) 
						( (actualField[posY][posX] & F_TYPE_MASK) | F_TELEPORT);
					}
					teleportStarted=false;
				}
				else if(c == KeyEvent.VK_UP && isCellAccessible(posX+1, posY)) {
					posX++;
					level[0].teleport[teleportNr].toX= posX;
				}
				else if(c == KeyEvent.VK_DOWN && isCellAccessible(posX-1, posY)) {
					posX--;
					level[0].teleport[teleportNr].toX= posX;
				}
				else if(c == KeyEvent.VK_RIGHT && isCellAccessible(posX, posY+1)) {
					posY++;
					level[0].teleport[teleportNr].toY= posY;
				}
				else if(c == KeyEvent.VK_LEFT && isCellAccessible(posX, posY-1)) {
					posY--;
					level[0].teleport[teleportNr].toY= posY;
				}
				repaint();
				return;
			}

			if (c == KeyEvent.VK_B && !trailStarted) { // beginn with setting of the field
				trailStarted=true;
				lastMove= 0;
				updateField();
				startX= posX;
				startY= posY;
				repaint();
			}
			else if (c == KeyEvent.VK_UP) {
				if (trailStarted) {
					isStartCell= false;
					if  (goUp(false) )
						solution += "U";
				}
				else if (isCellAccessible(posX+1, posY)){
					posX++;
					repaint();
				}
			}
			else if (c == KeyEvent.VK_DOWN) {
				if (trailStarted) {
					isStartCell= false;
					if (goDown(false) )
						solution += "D";
				}
				else if (isCellAccessible(posX-1, posY)){
					posX--;
					repaint();
				}
			}
			else if (c == KeyEvent.VK_LEFT)	{
				if (trailStarted) {
					isStartCell= false;
					if (goLeft(false))
						solution+= "L";
				}
				else if (isCellAccessible(posX, posY-1)){
					posY--;
					repaint();
				}
			}
			else if (c == KeyEvent.VK_RIGHT) {
				if (trailStarted) {
					isStartCell= false;
					if (goRight(false))
						solution += "R";
				}
				else if (isCellAccessible(posX, posY+1)){
					posY++;
					repaint();
				}
			}
			else if (c == KeyEvent.VK_P && trailStarted) {
				actualField[posY][posX]= (byte) (F_STATIC | flag);
				repaint();
			}
			else if (c == KeyEvent.VK_S && !gameStoped) {
				saveLevel();
			}


			// flags can only be set on clear F_ONE cells to avoid inconsistencies
			// also on the start-cell no flags are allowed
			else if (type == F_ONE && flag == 0 &&!isStartCell) {

				// *** debug ****
				// test for numbers (assumes that the keykodes for number-keys
				// are ordered incremetal: 
				// VK_xx = VK_0 + xx
				// VK_NUMPADxx is used for forced direction
				if (	(c >= KeyEvent.VK_1 && c <= KeyEvent.VK_9 ) ) {
					int nr= c- KeyEvent.VK_0;
					if (nr<1 || nr >9) {
						nr= c- KeyEvent.VK_NUMPAD0;
					}

					// debug
					System.out.printf("Number:%d NextNr:%d\n", nr, nextNr);
					
					
					if (nr == nextNr +1) {
						nextNr++;
					}
					if (nr == nextNr && nr >= 1 && nr <= 9) {
						actualField[posY][posX]= (byte)( type | F_NUMBER + nextNr);
						repaint();						
					}
				}
				else if(c == KeyEvent.VK_T && teleportNr < MAX_TELEPORTS) {
					teleportStarted=true;
					showTeleport[teleportNr]= true;
					level[0].teleport[teleportNr].fromX= posX;
					level[0].teleport[teleportNr].fromY= posY;
					level[0].teleport[teleportNr].toX= posX;
					level[0].teleport[teleportNr].toY= posY;
					level[0].nrTeleports= teleportNr+1;
					actualField[posY][posX]= (byte)( type | F_TELEPORT);
				}
				else if(c == KeyEvent.VK_I) {
					if (lastMove == F_DIR_UP && isCellAccessible(posX+1,posY)) {
						actualField[posY][posX]= (byte)( type | F_ICE);
						goUp(false);
					}
					else if (lastMove == F_DIR_DOWN && isCellAccessible(posX-1,posY)) {
						actualField[posY][posX]= (byte)( type | F_ICE);
						goDown(false);
					}
					else if (lastMove == F_DIR_LEFT && isCellAccessible(posX,posY-1)) {
						actualField[posY][posX]= (byte)( type | F_ICE);
						goLeft(false);
					}
					else if (lastMove == F_DIR_RIGHT && isCellAccessible(posX,posY+1)) {
						actualField[posY][posX]= (byte)( type | F_ICE);
						goRight(false);
					}
				}
				else if (c == KeyEvent.VK_U || c == KeyEvent.VK_NUMPAD8) {
					if (isCellAccessible(posX+1,posY)) {
						if (actualField[posY][posX+1] == (byte)(F_ONE | F_DIR_DOWN) ) {
							return;
						}
						actualField[posY][posX]= (byte)( type | F_DIR_UP);
						goUp(false);
					}
				}

				else if (c == KeyEvent.VK_D || c == KeyEvent.VK_NUMPAD2
								|| c == KeyEvent.VK_NUMPAD5) {
					if (isCellAccessible(posX-1,posY)) {
						if (actualField[posY][posX-1] == (byte)(F_ONE | F_DIR_UP) ) {
							return;
						}
						actualField[posY][posX]= (byte)( type | F_DIR_DOWN);
						goDown(false);
					}
				}
				else if (c == KeyEvent.VK_L || c == KeyEvent.VK_NUMPAD4) {
					if (isCellAccessible(posX,posY-1)) {
						if (actualField[posY-1][posX] == (byte)(F_ONE | F_DIR_RIGHT) ) {
							return;
						}					
						actualField[posY][posX]= (byte)( type | F_DIR_LEFT);
						goLeft(false);
					}
				}
				else if (c == KeyEvent.VK_R || c == KeyEvent.VK_NUMPAD6) {
					if (isCellAccessible(posX,posY+1)) {
						if (actualField[posY+1][posX] == (byte)(F_ONE | F_DIR_LEFT) ) {
							return;
						}					
						actualField[posY][posX]= (byte)( type | F_DIR_RIGHT);
						goRight(false);
					}
				}
			}

			//System.out.print("keyPressed:" + (int)c);			
			isKeyReleased=false;
		}
	}	

	protected boolean isCellAccessible(int x, int y) {

		if (x>=0 && y>=0 && x<F_SIZE_X && y<F_SIZE_Y &&
				x+y>= F_SIZE_CLIP1 && x+y <= F_SIZE_X-1 +F_SIZE_Y-1 -F_SIZE_CLIP2) {
			return true;
		}
		return false;
	}

	//	updateField is called in goUp/goDown/goRight/goLeft before posX and
//	posY is changed (to clear/update the cell, after the figure is moved)
//	in the editor, the cell, where the figure is, should be updated
//	to get the actual cell, lastMove stores the actual direction from posX,posY
	protected void updateField() {
		int type, flag;		
		int x= posX;
		int y= posY;
		if (!teleportStarted && !waitForTeleport) {
			switch (lastMove) {
			case F_DIR_RIGHT:
				y++;
				break;
			case F_DIR_LEFT:
				y--;
				break;
			case F_DIR_UP:
				x++;
				break;
			case F_DIR_DOWN:
				x--;
				break;
			}		
		}
		if (isCellAccessible(x,y)) {
			type= actualField[y][x] & F_TYPE_MASK;
			flag= actualField[y][x] & F_FLAG_MASK;
			if (type == F_EMPTY) {
				actualField[y][x]= F_ONE;
			}		
			else if (type == F_ONE) {
				actualField[y][x]= (byte) (F_TWO | flag);
			}		
			else if (type == F_TWO) {
				actualField[y][x]= (byte) (F_THREE | flag);
			}		
			else if (type == F_THREE) {
				actualField[y][x]= (byte) (F_STATIC | flag);			
			}		
		}
		// forced move with ice out of playfield may occur
		else {
			gameOver();
		}
	}
	protected void doTeleport() {
		int i;
		boolean teleportFound= false;
		System.out.println("Teleport");

		// search for teleport
		for (i=0; i<level[levelNr].nrTeleports; i++) {
			System.out.println(level[levelNr].teleport[i].fromX +" "+
					level[levelNr].teleport[i].fromY +" "+
					level[levelNr].teleport[i].toX +" "+
					level[levelNr].teleport[i].toY +" ");
			if (posX == level[levelNr].teleport[i].fromX &&
					posY == level[levelNr].teleport[i].fromY) {
				posX= level[levelNr].teleport[i].toX;
				posY= level[levelNr].teleport[i].toY;
				teleportFound=true;
				System.out.println("teleport found");
				break;
			}
		} // for (nr...
		if (teleportFound) {
			teleportStarted= true;
			updateField();
			teleportStarted= false;
		}		
		repaint();
	}

	// save level in Hexadecimal format
	// appends level to an existing level-file
	// or generates a new one, if a new filename is selected
	protected void saveLevel() {
		int returnVal;
		int x,y,i;
		String s;
		
		fc.resetChoosableFileFilters();
		fc.addChoosableFileFilter(new LevelFileFilter());
		fc.setAcceptAllFileFilterUsed(false);
		returnVal= fc.showSaveDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			String fileName = file.getPath();
			if (!fileName.toLowerCase().endsWith(".level")) {
				fileName += ".level";
			}
			System.out.print("File-Name: "+fileName+"\n");
			// make last cell static
			actualField[posY][posX]=
				(byte) ( (actualField[posY][posX] & F_FLAG_MASK)| F_STATIC);
			try {
				/* old version  -> new version uses teleport-flag on up&down
				// delete teleport_up- flags on teleport_down cells
				for (i=0; i<level[0].nrTeleports; i++) {
					actualField[level[0].teleport[i].toY]
					            [level[0].teleport[i].toX] &= F_TYPE_MASK;
				}
				//*/
				RandomAccessFile raFile= new RandomAccessFile(fileName,"rw");
				//count numbers of level-files
				int nextLevel=1;
				while(raFile.getFilePointer() < raFile.length()) {
					s= raFile.readLine();
					if (s != null) {
						if (s.length() >= 5) {
							s= s.substring(0, 5);
						}
						if ("Level".equals(s)) {
							nextLevel++;
						}
					}
					
				}
				raFile.seek(raFile.length()); // append to existing file
				raFile.writeBytes("Level "+Integer.toString(nextLevel)+"\n");
				raFile.writeBytes(toHexString(startX," "));
				raFile.writeBytes(toHexString(startY,"\n"));
				for (y=0; y<F_SIZE_Y; y++) {
					for (x=0; x<F_SIZE_X-1; x++) {
						raFile.writeBytes(toHexString(actualField[y][x]," "));
					}
					//raFile.write(actualField[y]);
					raFile.writeBytes(toHexString(actualField[y][x],"\n"));
				}
				// no teleports saved yet
				raFile.writeBytes(toHexString(level[0].nrTeleports,"\n"));
				for (i=0; i<level[0].nrTeleports; i++) {
					raFile.writeBytes(toHexString(level[0].teleport[i].fromX," "));
					raFile.writeBytes(toHexString(level[0].teleport[i].fromY," "));
					raFile.writeBytes(toHexString(level[0].teleport[i].toX," "));
					raFile.writeBytes(toHexString(level[0].teleport[i].toY,"\n"));
				}
				raFile.close();		
				fileName += ".solution.txt";
				raFile= new RandomAccessFile(fileName,"rw");
				raFile.seek(raFile.length()); // append to existing file
				raFile.writeBytes("Level "+Integer.toString(nextLevel)+"\r\n");
				//raFile.writeBytes(solution);
				//raFile.writeBytes("\r\n------------------------------\r\n");
				raFile.writeBytes(compactSolution(solution));
				raFile.writeBytes("\r\n------------------------------\r\n");
				raFile.close();		
				//
			} catch (IOException ioe) {
				System.out.printf("Error Writing File\n");
				return;
				// debug teleport-down cells should be reset with teleport-flag
			}
		}
		//	key-released keyevent is often not send because of the save-dialog
		//  therefore set the key-released flag manually
		isKeyReleased= true;  
		resetLevel();
	}
	/*
	 * 
	 */
	private String toHexString(int i, String space) {
		String s;
		s= Integer.toHexString(i);
		if (s.length()< 2 ) {
			s= "0"+s;
		}	
		return s+space;
	}

	private String compactSolution(String s){
		int count;
		String result="";

		if (s.length() <= 1) {
			return s;
		}
		count=1;
		for (int i=1; i<s.length(); i++) {
			if (s.charAt(i-1) == s.charAt(i)) {
				count++;
			}
			else {
				if (count > 1)  {
					result +=  String.valueOf(count);
				}
				result += s.charAt(i-1)+ " ";
				count=1;
			}
		}
		if (count > 1)
			result +=  String.valueOf(count);
		result += s.charAt(s.length()-1)+ "\r\n";
		return result;
	}

	// draws time
	protected void paintUpperLeftCorner(Graphics g) {
		final String strings[]= { 
				"Cursor-Keys: move figure",
				"B: begin trail",
				"P: permanent field",
				"I: Iceflag",
				"U/D/L/R, NUMPAD<8-2/5-4-6>: forced direction",
				"1-9: number flag (not numpad)",
				"T: make teleport",
				"S: save level",
				"N: new level"
		};
		int i;
		g.setFont(new Font("Sans-serif",Font.PLAIN,13));
		for (i=0; i<strings.length; i++) {
			g.setColor(Color.darkGray);
			g.drawString(strings[i], 5, 16 + i*15);
			g.setColor(Color.white);
			g.drawString(strings[i], 4, 15 + i*15);
		}
	}
	protected void paintLowerLeftCorner(Graphics g) {
	}
	protected void updateNextNr(int cell) {
		// do nothing
	}
} // end class


class LevelFileFilter extends FileFilter {

	public boolean accept(File file) {
		String filename = file.getName().toLowerCase();
		return filename.endsWith(".level") || file.isDirectory();
	}
	public String getDescription() {
		return "Level Files";
	}
}

