/*	Playfield.java
     
   Copyright 2007 Manfred Kopp
   
    This software distributed under the terms of the 
    GNU General Public License Version 3 (GPLv3) of the License, 
    or (at your option) any later version.
       
    This file is part of the game Step-By-Step.

    Step-By-Step is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Step-By-Step is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.step_by_step;

//import java.awt.Canvas;
//import java.awt.image.VolatileImage;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.io.IOException;
import java.io.File;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
//import java.awt.GraphicsConfiguration;
//import java.awt.GraphicsEnvironment;
//import java.awt.GraphicsDevice;
import java.awt.Color;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.Timer;
import javax.swing.JComponent;
//import java.awt.image.ColorConvertOp;

import java.awt.Font;
import java.io.RandomAccessFile; 
import java.applet.Applet; // needed for sound
import java.applet.AudioClip; // needed for sound
//import javax.sound.*;


/*
// debug
import java.awt.Rectangle;
//*/
/**
 * @author Manfred Kopp
 *
 */
public class Playfield extends JComponent 
implements KeyListener, ActionListener {
	static final long serialVersionUID= 123456123L; 


	// local classes	
	protected class Teleport {
		public int fromX;
		public int fromY;
		public int toX;
		public int toY;				
	};

	protected class Level {
		public int startX;
		public int startY;
		public byte field[][];
		public Teleport teleport[];
		public int nrTeleports; 
		Level(){
			field= new byte[F_SIZE_Y][F_SIZE_X];
			teleport= new Teleport[MAX_TELEPORTS];
			for (int i=0; i<MAX_TELEPORTS; i++)
				teleport[i]= new Teleport();
		}
	};	 
	protected final static int P_DY= -128; // Playfield deltaX & deltaY
	protected final static int P_DX=  275;


	protected final static byte F_DELTA_TYPE= 0x10; // increment for type static, one, two, three 
	protected final static byte F_LAST_TYPE=  0x30;  // last type

	protected final static byte F_ONE= 	 	0x00;
	protected final static byte F_TWO= 	 	0x10;
	protected final static byte F_THREE=  	0x20;
	protected final static byte F_STATIC= 	0x30;
	protected final static byte F_EMPTY = 	0x40; // Empty has no flags

	// flags for the 4 types F_ONE,F_TWO,F_THREE & F_STATIC
	protected final static byte F_NUMBER= 		0; //number 1..9 corresponds to 1..9

	protected final static byte F_DIR_INVERS=	0x1;	// flag to invert left/right and up/down
	protected final static byte F_DIR_UP= 		0xa;
	protected final static byte F_DIR_DOWN= 	F_DIR_UP ^ F_DIR_INVERS; // 0xb
	protected final static byte F_DIR_LEFT= 	0xc;
	protected final static byte F_DIR_RIGHT= 	F_DIR_LEFT ^ F_DIR_INVERS; // 0xe

	protected final static byte F_ICE= 				0xe;
	protected final static byte F_TELEPORT= 		0xf;  // teleport up 
	protected final static byte F_LASTFLAG= 		0xf;
	protected final static byte F_FLAG_MASK= 		0x0f;
	protected final static byte F_TYPE_MASK= 		0x70;

	protected final static int F_SIZE_X= 17;  //size of the field in x-direction
	protected final static int F_SIZE_Y= 17;  //size of the field in y-direction
	protected final static int F_SIZE_CLIP1= 4;  //clipping of cells on the left side
	protected final static int F_SIZE_CLIP2= 5;  //clipping of the cells on the right side

	protected final static int MAX_TELEPORTS= 5;
	protected final static int T_UP= 0;
	protected final static int T_DOWN= 1;

	protected int nrTeleports;

	protected final static long NO_HISCORE= -1;
	protected final static String LEVEL_FILENAME= "default.level"; 
	protected final static String HISCORE_SUFFIX= ".hiscore"; 

	protected final static int NR_NORMAL = 0;
	protected final static int NR_HILIGHT = 1;
	
	
	protected final static int MAX_NUM_LEVELS= 99;	// maximal number of levels	
	protected int levelNr, nrLevels, showLevelNr;
	protected int startX;
	protected int startY;
	protected int numberCount[] = new int[10];

	protected Level level[];
	protected long hiscore[]; 
	protected Teleport teleport[][];

	protected boolean isKeyReleased, gameStoped, gameOverFlag, falling; 
	protected boolean waitForAnyInput, waitForInput;
	protected int lastMove; // last move is used for cells with ice
	protected boolean waitForTeleport; 
	protected boolean levelSelectMode, helpShowMode;
	protected int levelSelect;
	protected byte actualField[][] = new byte[F_SIZE_Y][F_SIZE_X];
	protected boolean showTeleport[] = new boolean[MAX_TELEPORTS];
	protected int posX, posY, newX, newY;
	protected int nextNr;
	protected int cellsLeft;	
	protected int fallingPosX, fallingPosY, textPosX, textPosY;
	protected final static int fallingIncY=16;
	protected final static int textIncY=10;

	protected int step=0;
	protected boolean impossible_move = false;

	protected BufferedImage bImg[]= new BufferedImage[65];
	protected BufferedImage imgGameOver, imgLevelSolved, imgGameEnded;
	protected BufferedImage imgBackground, imgLevelSelection, imgHelp;
	protected BufferedImage imgLevelSolvedMark, imgPlate, imgCursorLayout;
	protected BufferedImage tImg[][]= new BufferedImage[2][MAX_TELEPORTS];
	protected BufferedImage shoeImg[][]= new BufferedImage[4][4];
	protected BufferedImage numberImg[][] = new BufferedImage[10][2];

	

	protected long lastTime;
	protected Timer timer, endGameTimer, moveTimer, impossibleMoveTimer;
	protected long startTime;
	protected long delayTime; // used to stop time when showing help screen

	protected AudioClip sfxStep1, sfxStep2;
	protected AudioClip sfxWarning, sfxLevelSolved, sfxGameOver, sfxTeleport;

	/*
	 *  Constructor
	 */
	public Playfield(){
		int i;
		loadImages();
		loadSounds();
		this.addKeyListener(this);
		isKeyReleased=true;
		timer= new Timer(1000, this);
		endGameTimer= new Timer(30, this); // corresponds to 33 fps
		//moveTimer= new Timer(60, this); 
		moveTimer= new Timer(90, this);		// corresponds to tap sound (2taps= 180ms) 

		impossibleMoveTimer = new Timer(500,this);
		
		hiscore= new long[MAX_NUM_LEVELS];
		level= new Level[MAX_NUM_LEVELS];
		for (i= 0; i< MAX_NUM_LEVELS; i++) {
			level[i]= new Level();
			hiscore[i]= NO_HISCORE;  
		}

		loadLevels();
		levelNr=0;
		
		// start with first level not solved
		while((hiscore[levelNr] != NO_HISCORE) && levelNr <nrLevels) {
			levelNr++;
		}
		if (levelNr >= nrLevels) {
			levelNr=0;
		}
		
		resetLevel();
		fallingPosX= fallingPosY= 0;
		textPosX= textPosY=0;
		//System.out.println("Constructor Playfield");
	}

	// initialisation after the Playfield is generated
	public void init() {

		//System.out.println("init");		
	}

	protected void loadLevels() {
		loadLevels(LEVEL_FILENAME);
		//loadLevels("test.level");
	}

	protected void loadLevels(String fileName) {
		RandomAccessFile raFile;
		File f;
		int i=0,y=0,x=0, nr;
		long fileLength;
		boolean eofReached;

		try {
			String s,subStr;
			f= new File(fileName);
			raFile= new RandomAccessFile(f,"r");
			fileLength= raFile.length();
			levelNr=0;
			nrLevels=0;
			eofReached= false;
			for (i=0; i< MAX_NUM_LEVELS; i++) {	
				do {
					s= raFile.readLine(); // skip String "Level xx\n"
					eofReached= (fileLength - raFile.getFilePointer() ) <= 0L;
					/*
					// debug
					System.out.printf("Level:%d <%s>\tpos:%d\t \tfilelength:%d\n",
							i, s, raFile.getFilePointer(),fileLength);
					System.out.print("eofReached:"+eofReached+"\n");
					//*/
					subStr=null;
					if (s != null) {
						if (s.length() >= 5) {
							subStr= s.substring(0, 5);
						}
					}
					/* debug
					System.out.print("s:"+s+" substr:"+subStr+"\n");
					
					boolean b1= !eofReached;
					boolean b2= ! "Level".equals(subStr);
					boolean b3= b1 && b2;
					System.out.print("b1:"+b1+ "\tb2:"+b2+"\tb3:"+b3+"\n");
					//*/
				}
				while ( (! "Level".equals(subStr) ) && !eofReached);
				if (! eofReached) {
					level[i].startX= readHexByte(raFile);
					level[i].startY= readHexByte(raFile);
					for (y= 0; y< F_SIZE_Y; y++) {
						for (x=0; x< F_SIZE_X; x++) {
							level[i].field[y][x]= (byte) readHexByte(raFile);
						}
					}
					// 0 <= nr <= MAX_TELEPORTS 
					nr= readHexByte(raFile);				
					level[i].nrTeleports =nr;				
					while (nr>0) {
						nr--;
						level[i].teleport[nr].fromX=  readHexByte(raFile);	
						level[i].teleport[nr].fromY=  readHexByte(raFile);	
						level[i].teleport[nr].toX= x= readHexByte(raFile);	
						level[i].teleport[nr].toY= y= readHexByte(raFile);
						// be compatible with old version, where the toX/toY
						// cells are not set to with the F_TELEPORT- flag
						level[i].field[y][x] &= F_TYPE_MASK;
						level[i].field[y][x] |= F_TELEPORT;
					}
					nrLevels= i+1;   // increment nrLevels only, if no error occured in level
				} // end if (!eofReached)
				if (eofReached) {
					break; // exit for-loop if eof is reached
				}
			} // end for (i= ..
			raFile.close();

		}
		catch (Exception e) {
			e.printStackTrace();// debug
			System.out.print("Error reading level-file in level: "+(i+1)+"\n");
			System.out.print("      remaining levels are not loaded\n");
		}
		if (nrLevels == 0) {
			System.err.print("Level-File corrupted: no level loaded:\n"+i+","+y+"\n");
			System.exit(-1);
		}
		// load HiScore-File
		fileName += HISCORE_SUFFIX;
		try {
			f= new File (fileName);
			raFile= new RandomAccessFile(f,"r");
			if (f.canRead()) {				
				for(i=0; i<nrLevels; i++){
					hiscore[i]= raFile.readLong();
				}
			}
			raFile.close();
		}
		catch (Exception e) {			
		}			
	}

	protected boolean writeHiscore() {
		return writeHiscore(LEVEL_FILENAME+ HISCORE_SUFFIX);
	}


	protected boolean writeHiscore(String fileName) {
		int i;
		RandomAccessFile raFile;
		File f;
		boolean ok= true;
		//System.err.println("write HiscoreFile:"+fileName+"\n");
		try {
			f= new File (fileName);
			//System.err.println(1);
			raFile= new RandomAccessFile(f,"rw");
			//System.err.println(2);
			ok=f.canWrite();
			//System.err.println(ok);
			if (ok) {				
				for(i=0; i<nrLevels; i++){
					raFile.writeLong(hiscore[i]);
					//System.err.println(i+":"+hiscore[i]+"\n");
				}
			}
			raFile.close();
			//System.err.println(5);
		}
		catch (Exception e) {
			ok= false;
		}			
		return ok;		
	}

	private int readHexByte(RandomAccessFile f) throws Exception{
		//int dummy;
		byte hexString[]= new byte[2];
		f.read(hexString);
		f.skipBytes(1); // skip ' ' or '\n' after hexByte 
		return Integer.parseInt(new String(hexString),16);
	}



	protected void resetField() {
		int nr,x,y;
		int type;

		if (levelNr>= nrLevels) {
			levelNr= nrLevels-1;
		}
		startX= level[levelNr].startX;
		startY= level[levelNr].startY;
		cellsLeft= 0;

		for (nr=1; nr<=9; nr++) {
			numberCount[nr]=0;
		}
		for (y= 0; y< F_SIZE_Y; y++) {
			System.arraycopy(level[levelNr].field[y],0,actualField[y],0, F_SIZE_X);
			for (x=0; x< F_SIZE_X; x++) {
				type= actualField[y][x] & F_TYPE_MASK;
				nr= (actualField[y][x] & F_FLAG_MASK)  - F_NUMBER;
				if (nr >= 1 && nr<=9) {
					numberCount[nr]++;
				}
				if (type==F_ONE || type==F_TWO || type==F_THREE){
					cellsLeft++;
				}
			}
		}		
		
		/*// debug
		for (nr=1; nr<=9; nr++) System.out.printf("%d:%d\t",nr,numberCount[nr]);
		//*/
		posX= startX;
		posY= startY;
		nextNr= 1;

		//*/	
	}


	protected void resetLevel() {
		resetField();
		startTime= System.currentTimeMillis();
		lastTime= startTime;
		timer.start();
		endGameTimer.stop();
		moveTimer.stop();
		step=0;
		lastMove= 0; // no last move
		waitForTeleport= false;
		gameStoped= false;
		falling= false;
		waitForAnyInput= false;
		levelSelectMode= false;
		helpShowMode= false;

		step=0;
		waitForInput= true;
		gameOverFlag= false;
		showLevelNr= levelNr;
		for (int i=0; i<MAX_TELEPORTS;i++) {
			showTeleport[i]=false;				
		}
		for (int i=0; i<level[levelNr].nrTeleports;i++) {
			showTeleport[i]=true;				
		}
		// reset hilights of numbers 2-9
		for (int i=2; i<=9; i++)
		{
			Graphics g;
			for (int type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_NUMBER + i].getGraphics();
				g.drawImage(bImg[type], 0, 0, this);       	
				g.drawImage(numberImg[i][NR_NORMAL], 0, 0, this);       	
			}

		}
	}

	protected void loadImages() 
	{ 
		BufferedImage tempImg;
		Graphics g;
		int i,type;

		bImg[F_EMPTY]= null;
		try 
		{
			// ground tiles
			
			bImg[F_STATIC] = ImageIO.read( new File 
					("gfx","tile-static.gif"));
			bImg[F_ONE] = ImageIO.read( new File 
					("gfx","tile-one.gif"));
			bImg[F_TWO] = ImageIO.read( new File 
					("gfx","tile-two.gif"));
			bImg[F_THREE]=ImageIO.read( new File 
					("gfx","tile-three.gif"));


			// shoe images
			
			for (i=0; i< 4; i++) {
				shoeImg[0][i]=
					ImageIO.read( new File 
							("gfx","figure-up_"+Integer.toString(i)+".png"));
				shoeImg[1][i]=
					ImageIO.read( new File 
							("gfx","figure-down_"+Integer.toString(i)+".png"));
				shoeImg[2][i]=
					ImageIO.read( new File 
							("gfx","figure-left_"+Integer.toString(i)+".png"));
				shoeImg[3][i]=
					ImageIO.read( new File 
							("gfx","figure-right_"+Integer.toString(i)+".png"));
			}
						
			
			imgGameOver= ImageIO.read( new File ("gfx","game_over.gif"));
			imgLevelSolved= ImageIO.read( new File ("gfx","level_solved.gif"));
			imgGameEnded= null;
			imgBackground= ImageIO.read( new File ("gfx","background.gif"));
			imgLevelSelection= ImageIO.read( new File ("gfx","level-selection.gif"));
			imgLevelSolvedMark= 
				ImageIO.read( new File ("gfx","level-solved-mark.gif"));
			imgHelp= 
				ImageIO.read( new File ("gfx","instructions.gif"));
			imgPlate= ImageIO.read( new File ("gfx","plate.gif"));
			imgCursorLayout= ImageIO.read( new File ("gfx","cursor_layout.gif"));

			for (i= 1; i<= F_LASTFLAG; i++) {
				for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
					bImg[type + i]= 
						new BufferedImage(64,32,BufferedImage.TYPE_4BYTE_ABGR_PRE);
					bImg[type + i].getGraphics().drawImage(bImg[type], 0, 0, this);
				}
			}
			// ice flags
			bImg[F_STATIC+ F_ICE] = ImageIO.read( new File 
					("gfx","ice-static.gif"));
			bImg[F_ONE+ F_ICE] = ImageIO.read( new File 
					("gfx","ice-one.gif"));
			bImg[F_TWO+ F_ICE] = ImageIO.read( new File 
					("gfx","ice-two.gif"));
			bImg[F_THREE+ F_ICE]=ImageIO.read( new File 
					("gfx","ice-three.gif"));

			//System.err.println("property:"+System.getProperty("sun.java2d.translaccel"));

			// number flags 

			for (i=1; i<=9; i++) {
				numberImg[i][NR_NORMAL]=  ImageIO.read(new File ("gfx",Integer.toString(i)+".png"));
				numberImg[i][NR_HILIGHT]= ImageIO.read(new File ("gfx",Integer.toString(i)+"-hilight.png"));
				for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
					g= bImg[type+ F_NUMBER + i].getGraphics();
					g.drawImage(numberImg[i][NR_NORMAL], 0, 0, this);       	
				}
			}
			// hilight number 1
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_NUMBER + 1].getGraphics();
				g.drawImage(bImg[type], 0, 0, this);       	
				g.drawImage(numberImg[1][NR_HILIGHT], 0, 0, this);       	
			}

			// teleport 
			// teleport-tile
			tempImg= ImageIO.read(new File ("gfx","teleport.gif"));
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+F_TELEPORT].getGraphics();
				g.drawImage(tempImg, 0, 0, this);       	
			}
			// teleport up/down arrow-images			
			for (i=0; i< MAX_TELEPORTS; i++) {

				tImg[T_UP][i]= ImageIO.read(
						new File("gfx","teleport-up_"+Integer.toString(i+1)+".png"));
				tImg[T_DOWN][i]= ImageIO.read(
						new File("gfx","teleport-down_"+Integer.toString(i+1)+".png"));
				//*/
				//tImg[T_UP][i]= tImg[T_DOWN][i]= null;
			}

			// directional flags
			tempImg= ImageIO.read(new File ("gfx","arrow-right.gif"));
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_DIR_RIGHT].getGraphics();
				g.drawImage(tempImg, 0, 0, this);       	
			}
			tempImg= ImageIO.read(new File ("gfx","arrow-up.gif"));
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_DIR_UP].getGraphics();
				g.drawImage(tempImg, 0, 0, this);       	
			}
			tempImg= ImageIO.read(new File ("gfx","arrow-down.gif"));
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_DIR_DOWN].getGraphics();
				g.drawImage(tempImg, 0, 0, this);       	
			}
			tempImg= ImageIO.read(new File ("gfx","arrow-left.gif"));
			for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
				g= bImg[type+ F_DIR_LEFT].getGraphics();
				g.drawImage(tempImg, 0, 0, this);       	
			}

			/*
			String typeString[]= {"one_","two_","three_","static_" }; 
			for (i= 1; i<= F_LASTFLAG; i++) {
				for (type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
					ImageIO.write(bImg[type + i], "png", 
							new File("gfx","tile_"+typeString[type/F_DELTA_TYPE]+
									Integer.toString(i)+".png"));
				}
			}
			//*/
		} 
		catch ( IOException e ) 
		{ 
			e.printStackTrace(); 
			System.exit(-1);
		} 
	} 	
	protected void loadSounds() 
	{ 
		File f1 = new File( "sfx","step.wav" );
		File f2 = new File( "sfx","level_solved.wav" );
		File f3 = new File( "sfx","game_over.wav" );
		File f4 = new File( "sfx","teleport.wav" );
		File f5 = new File( "sfx","warning.wav" );

		try {
			sfxStep1= 		Applet.newAudioClip( f1.toURL() );
			sfxStep2= 		Applet.newAudioClip( f1.toURL() );
			
			sfxLevelSolved= Applet.newAudioClip( f2.toURL() );
			sfxGameOver= 	Applet.newAudioClip( f3.toURL() );
			sfxTeleport= 	Applet.newAudioClip( f4.toURL() );
			sfxWarning= 	Applet.newAudioClip( f5.toURL() );
		}
		catch (Exception e) {
			// error reading soundfiles
			e.printStackTrace(); 
			System.exit(-1);
		}
	}

	protected long getSeconds(){
		if (!gameStoped) {
			lastTime= System.currentTimeMillis();
		}
		return (lastTime-startTime)/1000;
	}

//	draws plate with level, time best_time and tiles_left
	protected void paintUpperLeftCorner(Graphics g) {
		long min, sec;
		String string;

		g.drawImage(imgPlate, 0, 0, this);

		g.setFont(new Font("Monospaced",Font.BOLD,20));
		string= String.format("LEVEL %d", showLevelNr+1);
		g.setColor(Color.black);
		g.drawString(string, 18, 27);
		g.setColor(Color.white);
		g.drawString(string, 15, 25);


		sec= getSeconds();
		min= sec / 60 ;
		sec= sec % 60;
		string= String.format("Time  %02d:%02d", (int)min, (int)sec);
		g.setColor(Color.black);
		g.drawString(string, 23, 52);
		g.setColor(Color.white);
		g.drawString(string, 20, 50);

		if (hiscore[showLevelNr]== NO_HISCORE) {
			string= String.format("Best  -");
		}
		else {
			string= String.format("Best  %02d:%02d", 
					(int)hiscore[showLevelNr]/60, (int)hiscore[showLevelNr]%60);			
		}
		g.setColor(Color.black);
		g.drawString(string, 23, 72);
		g.setColor(Color.white);
		g.drawString(string, 20, 70);

		string= String.format("Tiles %02d", cellsLeft);
		g.setColor(Color.black);
		g.drawString(string, 23, 95);
		g.setColor(Color.white);
		g.drawString(string, 20, 93);

	}

	protected void paintUpperRightCorner(Graphics g) {
		int img_dx= (this.getWidth()- imgCursorLayout.getWidth()) - 10;
		int img_dy= +20;
		final String string= "Cursor Directions:"; 
		g.setFont(new Font("Sans-serif",Font.PLAIN,13));

		g.setColor(Color.black);
		g.drawString(string, img_dx+2, 16 );
		g.setColor(Color.white);
		g.drawString(string, img_dx, 15);

		g.drawImage(imgCursorLayout, img_dx, img_dy, this);		
	}

	protected void paintLowerLeftCorner(Graphics g) {
		final String strings[]= { 
				"Cursor-Keys... Move Figure",
				"F1 or H... Show Help",
				"F2 or R... Reset Level",
				"F3 or L... Select Level"
		};
		int dy= getHeight() - (strings.length-1) * 16 -5;
		g.setFont(new Font("Sans-serif",Font.PLAIN,13));

		for (int i=0; i<strings.length; i++) {
			g.setColor(Color.black);
			g.drawString(strings[i], 8, dy+2 + i*16);
			g.setColor(Color.white);
			g.drawString(strings[i], 7, dy   + i*16);
		}
	}

	protected void paintLowerRightCorner(Graphics g) {
	}

	public void	paint(Graphics g) {
		int x,y, dx,dy, cell,i;

		Graphics2D g2d= (Graphics2D)g;
		
		g2d.addRenderingHints(
		    new RenderingHints(RenderingHints.KEY_TEXT_ANTIALIASING, 
		    		RenderingHints.VALUE_TEXT_ANTIALIAS_ON));
		/* debug
		long deltaTime;
		Date tempDate= new Date();		
		deltaTime= tempDate.getTime();
		System.out.println("step:"+step);

		//*/
		dx=imgBackground.getWidth();
		dy=imgBackground.getHeight();

		for (x=0; x<getWidth(); x+=dx){
			for (y=0; y<getHeight();y+=dy){
				g.drawImage(imgBackground, x, y, this);
			}
		}

		if (levelSelectMode) {
			paintLevelSelection(g);
			return;
		}
		else if (helpShowMode) {
			paintHelpScreen(g);
			return;
		}


		paintUpperLeftCorner(g);
		paintUpperRightCorner(g);
		paintLowerLeftCorner(g);
		paintLowerRightCorner(g);

		/*
		// debug show clip rectangle
		Rectangle r= g.getClipBounds();
		g.drawRect(r.x, r.y, r.width-1, r.height-1);
		//*/
		
		if (posY< 0 && falling) {
			if (lastMove>= F_DIR_UP && lastMove<= F_DIR_RIGHT) {
				g.drawImage(shoeImg[lastMove-F_DIR_UP][0],fallingPosX,fallingPosY,this );
			}
			else { 					
				g.drawImage(shoeImg[0][0],fallingPosX,fallingPosY,this );
			}
		}
		for (y=0; y < F_SIZE_Y; y++) {
			for (x=F_SIZE_X -1; x>=0; x--) {
				if ((x+y>=F_SIZE_CLIP1) && (x+y <= F_SIZE_X-1 + F_SIZE_Y-1 -F_SIZE_CLIP2 )) {
					cell= actualField[y][x];
					g.drawImage(bImg[cell], P_DY+ 32*(x+y) ,P_DX+16*(y-x),this);
				}
			} // end for (x=...
			// draw figure (if falling
			if (posY==y && falling) {
				if (lastMove>= F_DIR_UP && lastMove<= F_DIR_RIGHT) {
					g.drawImage(shoeImg[lastMove-F_DIR_UP][0],fallingPosX,fallingPosY,this );
				}
				else { 					
					g.drawImage(shoeImg[0][0],fallingPosX,fallingPosY,this );
				}
				x= posX-1;
				if (x>= 0) {
					cell= actualField[y][x];
					g.drawImage(bImg[cell], P_DY+ 32*(x+y) ,P_DX+16*(y-x),this);
				}

			}			
		} // end for (y=...
		if (posY==F_SIZE_Y && falling) {
			if (lastMove>= F_DIR_UP && lastMove<= F_DIR_RIGHT) {
				g.drawImage(shoeImg[lastMove-F_DIR_UP][0],fallingPosX,fallingPosY,this );
			}
			else { 					
				g.drawImage(shoeImg[0][0],fallingPosX,fallingPosY,this );
			}
		}

		if (!falling) {	
			if (impossible_move)
			{
				//System.out.println("Paint1:"+System.currentTimeMillis());

				g.drawImage(shoeImg[lastMove-F_DIR_UP][3],
						P_DY+32*(posX+posY)-32,P_DX+16*(posY-posX)-16-32,this );
			
			}
			else if (lastMove>= F_DIR_UP && lastMove<= F_DIR_RIGHT) {
				g.drawImage(shoeImg[lastMove-F_DIR_UP][step],
						P_DY+32*(posX+posY)-32,P_DX+16*(posY-posX)-16-32,this );
			}
			else { 					
				g.drawImage(shoeImg[0][step],
						P_DY+32*(posX+posY)-32,P_DX+16*(posY-posX)-16-32,this );
			}
		}

		// draw teleports
		for (i=0; i< level[showLevelNr].nrTeleports; i++) {
			if (showTeleport[i]) {
				x= level[showLevelNr].teleport[i].fromX;
				y= level[showLevelNr].teleport[i].fromY; 
				g.drawImage(tImg[T_UP][i], P_DY+ 32*(x+y) ,P_DX+16*(y-x)-16,this);
				x= level[showLevelNr].teleport[i].toX;
				y= level[showLevelNr].teleport[i].toY; 				
				g.drawImage(tImg[T_DOWN][i], P_DY+ 32*(x+y) ,P_DX+16*(y-x)-16,this);				
			}
		}

		/*  debug show ms for painting
		tempDate= new Date();
		deltaTime= tempDate.getTime() - deltaTime;
		System.out.println((int)deltaTime);
		//*/
		if (gameStoped) {
			g.drawImage(imgGameEnded, textPosX, textPosY, this);
			if (waitForAnyInput)
			{
				final String continueString ="press any key to continue";
				x=253;
				y=380;
				g.setFont(new Font("Sans-serif",Font.PLAIN,26));
				g.getFontMetrics();
				g.setColor(Color.black);
				g.drawString(continueString, x+2,y+1);
				g.setColor(Color.white);
				g.drawString(continueString, x, y);
			}
		}
	}

	protected void paintHelpScreen(Graphics g) {
		int img_dx= (this.getWidth()- imgLevelSelection.getWidth())/2;
		int img_dy= (this.getHeight()- imgLevelSelection.getHeight())/2;

		g.drawImage(imgHelp, img_dx, img_dy, this);
	}

	protected void paintLevelSelection(Graphics g){
		int x,y, l,row, column;
		int levels_dx, levels_dy, row_dy, column_dx;
		int cursor_x, cursor_y;
		int img_dx= (this.getWidth()- imgLevelSelection.getWidth())/2;
		int img_dy= (this.getHeight()- imgLevelSelection.getHeight())/2;
		String str;

		g.drawImage(imgLevelSelection, img_dx, img_dy, this);

		levels_dx= 30 + img_dx;
		levels_dy= 55 + img_dy  +30; 

		column_dx= 60;
		row_dy=  38;
		cursor_x= levels_dx + ((levelSelect-1) %10 ) *column_dx - 14;
		cursor_y= levels_dy + ((levelSelect-1) /10 ) *row_dy    - 24;		
		g.drawImage(bImg[F_ONE], cursor_x, cursor_y, this);

		for (l= 1; l<= 99 ; l++){//nrLevels; l++) {
			row= (l-1)/10;
			column= (l-1) %10;
			x= levels_dx + column* column_dx;
			y= levels_dy + row* row_dy;
			g.setFont(new Font("Monospaced",Font.BOLD,30));
			str= String.format("%02d", l);
			g.setColor(Color.black);
			g.drawString(str, x+2,y+3);
			g.setColor(Color.white);
			if ( l> nrLevels){
				g.setColor(Color.gray);
			}
			g.drawString(str, x, y);
			if (hiscore[l-1] != NO_HISCORE) {
				g.drawImage(imgLevelSolvedMark, x+20, y-12, this);
			}
		}
	}


	public void keyTyped(KeyEvent e) {
		/*
		// key code is 0
		int c= e.getKeyCode();
		System.out.print("keyTyped:");
		System.out.println((int)c);
		//*/
	}
	private void SelectLevelKeyEvent(int c){
		if (c== KeyEvent.VK_ENTER){
			levelNr= levelSelect-1;
			levelSelectMode= false;
			resetLevel();
			repaint();
		}
		else if (c == KeyEvent.VK_UP) {
			if (levelSelect>10) {
				levelSelect -= 10;
				repaint();
			}
		}
		else if (c == KeyEvent.VK_DOWN) {
			if (levelSelect +10 <= nrLevels) {
				levelSelect += 10;
				repaint();
			}
		}
		else if (c == KeyEvent.VK_LEFT)	{
			if (levelSelect>1){
				levelSelect--;
				repaint();
			}
		}
		else if (c == KeyEvent.VK_RIGHT) {
			if (levelSelect <nrLevels ){
				levelSelect++;
				repaint();
			}
		} 
	}
	public void keyPressed(KeyEvent e) {
		int c= e.getKeyCode();
		/* debug
		System.out.print("keyPressed:");
		System.out.println((int)c);
		//*/

		if (isKeyReleased ) {
			if (levelSelectMode){
				SelectLevelKeyEvent(c);
				isKeyReleased=false;
				return;
			}		
			else if (helpShowMode) {
				// add delay time to start time
				delayTime= System.currentTimeMillis() - delayTime;
				startTime += delayTime;
				helpShowMode= false;
				isKeyReleased=false;
				repaint();
				return;
			}
			if (!gameStoped && waitForInput) {
				if (c == KeyEvent.VK_UP) {
					goUp(false);
				}
				else if (c == KeyEvent.VK_DOWN) {
					goDown(false);
				}
				else if (c == KeyEvent.VK_LEFT)	{
					goLeft(false);
				}
				else if (c == KeyEvent.VK_RIGHT) {
					goRight(false);
				}
			}
			if (c == KeyEvent.VK_R || c == KeyEvent.VK_F2 || waitForAnyInput) {
				resetLevel();
				repaint();
			}
			else if (c == KeyEvent.VK_L || c == KeyEvent.VK_F3) {
				levelSelectMode= true;
				levelSelect= levelNr+1;
				repaint();
			}
			else if (c == KeyEvent.VK_H || c == KeyEvent.VK_F1) {
				delayTime= System.currentTimeMillis();
				helpShowMode= true;
				repaint();
			}

			//System.out.print("keyPressed:" + (int)c);			
			isKeyReleased=false;
		}
	}	
	public void keyReleased(KeyEvent e) {		
		/* debug
		int c= e.getKeyCode();
		System.out.print("keyReleased:");
		System.out.println((int)c);
		//*/
		isKeyReleased= true;
	}	
	protected void updateField() {
		int type, flag;
		type= actualField[posY][posX] & F_TYPE_MASK;
		flag= actualField[posY][posX] & F_FLAG_MASK;

		if ((flag>= F_NUMBER +1 ) && (flag <= F_NUMBER+9)) {
			flag= 0; // delete flag
		}
		if (type == F_ONE) {
			actualField[posY][posX]= F_EMPTY;
			cellsLeft--;
		}		
		else if (type == F_TWO) {
			actualField[posY][posX]= (byte) (F_ONE | flag);
		}		
		else if (type == F_THREE) {
			actualField[posY][posX]= (byte) (F_TWO | flag);			
		}		
		// static type only change number-flag
		else if (type == F_STATIC) {
			actualField[posY][posX]= (byte)(F_STATIC | flag);			
		}		
	}
	protected void checkMoves() {
		boolean movesLeft= false;

		if (posX<0 || posX >= F_SIZE_X || posY<0  || posY>= F_SIZE_Y) {
			// a direction button or a ice button forced a move outside
			// the field
			gameOver();
			return;
		}
		else if (cellsLeft <= 0) {
			levelSolved();
			return;
		}
		else {
			if (posX < F_SIZE_X -1) {
				//System.out.println(posX);
				if (isCellAccessible( posX+1, posY)){
					movesLeft= true;
					
				}
			}
			if ((!movesLeft) && (posX > 0) ) {
				if (isCellAccessible(posX-1, posY)){
					movesLeft= true;
				}
			}
			if ((!movesLeft) && (posY < F_SIZE_Y -1)) {
				if (isCellAccessible(posX, posY+1)){
					movesLeft= true;
				}
			}
			if ((!movesLeft) && (posY > 0) ){
				if (isCellAccessible(posX, posY-1)){
					movesLeft= true;
				}
			}
			// check for teleport-up
			if ((actualField[posY][posX] & F_FLAG_MASK) == F_TELEPORT ) {
				// cell maybe teleport up or down
				if (isTeleportUp (posX,posY)) {	
					movesLeft= true;					
				}
			}
			if (!movesLeft) {
				/* old version - 
				if ( cellsLeft==1 && 
						(actualField[posY][posX] & F_TYPE_MASK) == F_ONE)  {
					// standing on last F_ONE-cell
					levelSolved();
				}
				else {
					gameOver();
				}
				//*/
				gameOver();
			} // end if(!movesLeft)
		} // end else (inside field)
	}

	protected void updateNextNr(int cell) {
		int nr;
		nr = (cell & F_FLAG_MASK) - F_NUMBER;
		if (nr == nextNr && nr>=1 && nr <= 9){
			numberCount[nr]--;
			// debug
			//System.out.printf("NumberCount[%d]=%d\n", nr,numberCount[nr]);
			if (numberCount[nr] <= 0){
				nextNr++;
				if (nextNr <= 9)
				{					
					Graphics g;
					for (int type=0; type <= F_LAST_TYPE; type+= F_DELTA_TYPE) {
						g= bImg[type+ F_NUMBER + nextNr].getGraphics();
						g.drawImage(bImg[type], 0, 0, this);       	
						g.drawImage(numberImg[nextNr][NR_HILIGHT], 0, 0, this);       	
					}
				}
			}
		}
	}		
	protected boolean isCellAccessible(int x, int y) {
		int cell,mask;
		cell= actualField[y][x];

		mask = (cell & F_FLAG_MASK);
		if (cell == F_EMPTY) { 
			return false;
		}
		/*
		else if (mask == (fromDirection^F_DIR_INVERS) ){
			return false;
		}
		//*/
		else if ( (mask >= 1+F_NUMBER) && (mask <= 9+F_NUMBER) 
				&& (mask != nextNr)) {
			return false;
		}
		else 
			return true;		
	}	
	protected void gameOver() {
		repaint();
		if (!gameStoped) {
			imgGameEnded= imgGameOver;
			gameEnded();
			if (! waitForTeleport) {
				sfxGameOver.play();
			}
		}

		// game may be Stoped and last cell has an direction or ice-flag
		// therefore a forced move can call gameOver() a second time
		//System.out.print("x,y:"+posX+","+posY);
		if (posX<0 || posX>=F_SIZE_X|| posY<0 || posY>=F_SIZE_Y) {
			falling= true;
		}
		else if(actualField[posY][posX]== F_EMPTY) {
			falling=true;
		}	
		if (falling) {
			fallingPosX= P_DY+ 32*(posX+posY)-32;
			fallingPosY= P_DX+ 16*(posY-posX)-16-32;
			//System.out.print("falling:"+fallingPosX+","+fallingPosY);
		}
	}

	protected void levelSolved() {
		if (!gameStoped) {
			//System.out.println("************  Level Solved **************");
			imgGameEnded= imgLevelSolved;
			if (( hiscore[levelNr]== NO_HISCORE) || 
					hiscore[levelNr]> getSeconds()){
				hiscore[levelNr]= getSeconds();
				writeHiscore();
			}
			gameEnded();			
			levelNr++;
			if (! waitForTeleport) {
				sfxLevelSolved.play();
			}
		}
	}	

//	common Routines for gameOver and levelSolved	
	private void gameEnded() {
		timer.stop();
		gameStoped= true;
		endGameTimer.start();		
		textPosY= -100;
		textPosX= (getWidth()- imgGameEnded.getWidth() )/2;	

		/*
		// debug:
		if (imgGameEnded == imgGameOver) {
			System.out.print("->Game Over");
		}
		else if (imgGameEnded == imgLevelSolved) {
			System.out.print("->LevelSolved");
		}
		else {
			System.out.print("-> ***** nothing");
		}
		//*/
	}
	// debug old comment
	/**
	 * checkForcedMoves(int cell) checks, if the actual field with content cell
	 * has a direction- or ice-flag. If yes, it makes a forced move and starts the
	 * forcedMoveTimer. After the elapsed time, the Movement is done by calling
	 * goUp() / goDown() / goRight() /goLeft().
	 * 
	 * @param cell contains type (F_ONE,F_TWO,F_THREE,F_STATIC) and flags (direction,ice,numbers)
	 */
	protected void checkForcedMoves(int x, int y) {	
		int deltaX,deltaY;
		int cell= actualField[y][x];
		int type= cell & F_FLAG_MASK;
		switch (type) {
		case F_DIR_UP:
			goUp(true);
			break;
		case F_DIR_DOWN:
			goDown(true);
			break;
		case F_DIR_RIGHT:
			goRight(true);
			break;
		case F_DIR_LEFT:
			goLeft(true);
			break;
		case F_TELEPORT:
			if ( isTeleportUp(x,y) ) {
				waitForInput= false;
				moveTimer.start();	// go in direction/teleport after time-triggers
				waitForTeleport= true;
			}
			break;
		case F_ICE:			
			deltaX= (lastMove==F_DIR_UP)   ?  1 : 0;
			deltaX= (lastMove==F_DIR_DOWN) ? -1 : deltaX;
			deltaY= (lastMove==F_DIR_RIGHT)?  1 : 0;
			deltaY= (lastMove==F_DIR_LEFT) ? -1 : deltaY;
			goInDirection(deltaX,deltaY, lastMove, true);			
			break;
		}			
	}

	private boolean goInDirection(int deltaX, int deltaY, 
			int move, boolean forcedMove){
		int x,y;
		x= posX+deltaX;
		y= posY+deltaY;

		if (x >= 0 && x< F_SIZE_X && y >= 0 && y < F_SIZE_Y) {
			int cell= actualField[y][x];
			if (isCellAccessible(x, y) ) {
				doStep(x,y,move);
				updateNextNr(cell);
				//checkMoves();
				//checkForcedMoves(cell);
				return true;
			}
			else if(forcedMove) {
				doStep(x,y,move);
				gameOverFlag=true;	
			}
		}
		else if (forcedMove) {
			doStep(x,y,move);
			gameOverFlag=true;	
		}
		if (!gameOverFlag)
		{
			this.sfxWarning.play();
			impossible_move = true;
			lastMove = move;
			//System.out.println("Start:"+System.currentTimeMillis());
			//repaint();+
			repaint(P_DY+32*(posX+posY)-32, P_DX+16*(posY-posX)-16-32,
					shoeImg[move - F_DIR_UP][3].getWidth(), 
					shoeImg[move - F_DIR_UP][3].getHeight());

			impossibleMoveTimer.restart();
			//impossibleMoveTimer.
		}
		return false;
	}

	private void doStep(int x, int y, int move) {
		lastMove= move;
		updateField();
		newX= x;
		newY= y;
		step=1;
		waitForInput= false; // dont accept input-keys during animation
		repaint();
		moveTimer.start();		
	}

	protected boolean goUp(boolean forcedMove) { 
		boolean ok= goInDirection(1,0, F_DIR_UP, forcedMove);
		return ok;
	}	
	protected boolean goDown(boolean forcedMove) { 
		boolean ok= goInDirection(-1,0, F_DIR_DOWN, forcedMove);
		return ok;
	}	
	protected boolean goRight(boolean forcedMove) { 
		boolean ok= goInDirection(0,1, F_DIR_RIGHT, forcedMove);
		return ok;
	}	
	protected boolean goLeft(boolean forcedMove) { 
		boolean ok= goInDirection(0,-1, F_DIR_LEFT, forcedMove);
		return ok;
	}	

	protected boolean isTeleportUp(int x, int y) {
		boolean result= false;
		for (int i=0; i<level[levelNr].nrTeleports; i++) {		
			if (	x == level[levelNr].teleport[i].fromX &&
					y == level[levelNr].teleport[i].fromY ) {
				result=true;
				break;
			}
		}
		return result;
	}
	
	protected void doTeleport() {
		int i, cell;
		//System.out.println("Teleport");
		sfxTeleport.play();
		//lastMove= F_TELEPORT;
		updateField();
		cell= actualField[posY][posX];
		// search for teleport
		for (i=0; i<level[levelNr].nrTeleports; i++) {
			/*
			System.out.println(level[levelNr].teleport[i].fromX +" "+
					level[levelNr].teleport[i].fromY +" "+
					level[levelNr].teleport[i].toX +" "+
					level[levelNr].teleport[i].toY +" ");
			//*/
			if (posX == level[levelNr].teleport[i].fromX &&
					posY == level[levelNr].teleport[i].fromY) {
				posX= level[levelNr].teleport[i].toX;
				posY= level[levelNr].teleport[i].toY;
				if (cell == F_EMPTY) {
					showTeleport[i]= false;
				}
				//System.out.println("teleport found");
				break;
			}
		} // for (nr...
		repaint();
		if (isCellAccessible(posX, posY) ){
			checkMoves();
		}
		else {
			gameOver();
		}
	}

	/**
	 *  Event-Listenter callback function for Timer event
	 */	

	public void	actionPerformed(ActionEvent e) {
		//System.out.print(e.paramString());
		//System.out.println(e.getActionCommand());


		if (e.getSource() == impossibleMoveTimer)
		{
			//System.out.println("Action performed:"+System.currentTimeMillis());

			impossible_move = false;
			impossibleMoveTimer.stop();
			repaint();
			return;
		}
		// do step-animation
		else if (e.getSource() == moveTimer && step!= 0) {
			if (step==1) 		sfxStep1.play();
			else if (step ==2) 	sfxStep2.play();
			step= (step+1)%3;			
			if (step == 0) {
				moveTimer.stop();
				posX= newX;
				posY= newY;
				waitForInput= true;
				if (gameOverFlag) {
					gameOverFlag= false;
					gameOver();
					return;
				}				
				checkMoves();
				checkForcedMoves(posX,posY);
			}
			repaint();
			return;

		}
		else if (e.getSource() == moveTimer && waitForTeleport) {
			moveTimer.stop();			
			waitForTeleport= false;			
			doTeleport();
			waitForInput= true;
			if (gameStoped) {
				if (imgGameEnded == imgLevelSolved) {
					sfxLevelSolved.play();
				}
				else {
					sfxGameOver.play();
				}
			}
		}

		else if (e.getSource() == timer) {
			//repaint();
			repaint(0,0,imgPlate.getWidth(),imgPlate.getHeight()); // cliping rectangle for time display
			return;
		}
		else { // game has stoped (game over or level solved
			//repaint(fallingPosX, fallingPosY-fallingIncY, 32, 32+fallingIncY); // falling Repaint
			//repaint(); //

			if (falling && fallingPosY < getHeight()) {
				fallingPosY += fallingIncY;
				// get img-Size of shoe
				BufferedImage imgTemp= shoeImg[0][0];
				if (lastMove>= F_DIR_UP && lastMove<= F_DIR_RIGHT) {
					imgTemp= shoeImg[lastMove-F_DIR_UP][0];
				}
				repaint(P_DY+32*(posX+posY)-32, fallingPosY - fallingIncY,
						imgTemp.getWidth(), imgTemp.getHeight() +fallingIncY );
			}
			/*
				System.err.println((this.getHeight()- imgGameEnded.getHeight())/2);
				System.err.println((this.getHeight()));
				System.err.println((imgGameEnded.getHeight()));
				//*/
			if (textPosY< (this.getHeight()- imgGameEnded.getHeight())/2) {
				textPosY+= textIncY;
				repaint((this.getWidth()- imgGameEnded.getWidth())/2 , textPosY- textIncY,
						imgGameEnded.getWidth(), imgGameEnded.getHeight() + textIncY);
			}
			else  {
				if (!falling || fallingPosY >= getHeight()) {
					//falling figure is out of view and 
					endGameTimer.stop();
					waitForAnyInput= true;
					repaint();
				}
			}
		}
	}
}
